local ADDON_NAME, ns = ...

local Printer = LibStub:GetLibrary("OakPrint-1.0")(addon_name)

local Data = ns.Data

local nameTable = ns.nameTable
local specTable = ns.specTable
local classTable = ns.classTable
local dataobj = ns.dataobj
local interruptTable = ns.interruptTable

local tContains, tinsert, CheckInteractDistance = tContains, tinsert, CheckInteractDistance

local function tremovebyval(t, val)
    for k, v in ipairs(t) do
        if v == val then
            tremove(t, k)
            break
        end
    end
end

--------------------------------------------------------------------------------
-- Inspection
--------------------------------------------------------------------------------

local notifyRegister = {}
local inspecting
local inspect_timeout

local shortcutFilter = {}
local specFilter = setmetatable({}, { __index = Data.irrelevantSpec })
-- default filter

local function RegisterInspection(name)
    Printer.debug(2, "Registering %s for inspection", name)
    if not tContains(notifyRegister, name) then
        tinsert(notifyRegister, name)
    end
    if not inspecting then
        Printer.debug(3, "Next inspection is for %s (calling NotifyInspect)", name)
        NotifyInspect(name)
        inspect_timeout = 4
        inspecting = UnitGUID(name)
    end
end

local function InspectNext()
    if next(notifyRegister) then
        RegisterInspection(notifyRegister[#notifyRegister])
    end
end

local function InspectGUID(guid)
    local name = nameTable[guid]
    assert(name, "No name associated with GUID " .. guid)
    Printer.debug(2, "Name: %s, GUID: %s", name, guid)

    local specID = GetInspectSpecialization(name)
    specTable[guid] = specID
    if specFilter[specID] then
        Printer.debug(2, "%s has irrelevant spec", name)
        interruptTable[guid] = nil --"UNREGISTER"
    else
        if not interruptTable[guid] then
            interruptTable[guid] = 0
        end
    end
    tremovebyval(notifyRegister, name)
    ClearInspectPlayer()
    inspecting = nil
end

--------------------------------------------------------------------------------
-- Polling
--------------------------------------------------------------------------------

local function PollByUnitID(unitID, force)
    if not UnitIsPlayer(unitID) then return end
    local name = UnitName(unitID)
    local guid = UnitGUID(unitID)
    if not guid or not name or name == "Unknown" then return end
    if force or not interruptTable[guid] then
        Printer.debug(2, "Processing %s:%s", name, guid)
        local class = select(2, UnitClass(unitID))
        if class == "WARLOCK" then
            Printer.debug(2, "%s is a warlock, we don't like warlocks")
            return
        end
        classTable[guid] = class
        if Data.spellOfClass[class] and not shortcutFilter[class] then
            -- all specs of the class have the same interrupt
            interruptTable[guid] = 0
        else
            RegisterInspection(name)
        end
    end
end

local function PurgeUnit(guid)
    local name = nameTable[guid]
    Printer.debug(3, "Purging %s:%s", name, guid)
    interruptTable[guid] = nil
    tremovebyval(notifyRegister, name)
end

local function PollGroup(force)
    Printer.debug(1, "Clearing stale data")
    for guid, _ in interruptTable() do
        local name = nameTable[guid]
        if force or not UnitInRaid(name) and not UnitInParty(name) then
            PurgeUnit(guid)
        end
    end
    if IsInGroup() then
        Printer.debug(1, "Polling group")
        local num = GetNumGroupMembers()
        local prefix

        if IsInRaid() then
            prefix = 'raid'
        else
            prefix = 'party'
            PollByUnitID('player')
            num = num - 1
        end

        for i=1,num do
            local unit = prefix .. i
            PollByUnitID(unit)
        end
    end
    inspect_timeout = 1
end

--------------------------------------------------------------------------------
-- Controller
--------------------------------------------------------------------------------

local controller = CreateFrame("Frame")

function controller.OnUpdate(self, elapsed)
    if inspecting then
        if inspect_timeout and inspect_timeout < elapsed then
            Printer.debug(3, "Inspection timed out, restarting...")
            ClearInspectPlayer()
            inspecting = nil
            InspectNext()
        else
            inspect_timeout = inspect_timeout - elapsed
        end
    end
    if #notifyRegister == 0 then
        Printer.debug(2, 'Nothing left to inspect, unregistering update script')
        controller:SetScript('OnUpdate', nil)
        dataobj.polling = false
    end
end

function controller:INSPECT_READY(guid)
    if inspecting == guid then
        InspectGUID(guid)
        InspectNext()
    else
        Printer.debug(3, "%s is not the current inspection target", UnitName(guid) or guid)
    end
end

function controller:GROUP_ROSTER_UPDATE()
    Printer.debug(2, "Roster Update")
    if IsInGroup() and not controller:GetScript('OnUpdate') then
        Printer.debug(3, "Registering update script")
        controller:SetScript('OnUpdate', controller.OnUpdate)
        dataobj.polling = true
    elseif not IsInGroup() then
        Printer.debug(3, "Left group, unregistering update script")
        controller:SetScript('OnUpdate', nil)
    end
    PollGroup()
end

function controller:PLAYER_SPECIALIZATION_CHANGED(unitID)
    if not unitID then return end
    Printer.debug(2, "%s changed spec", unitID)
    if UnitInRaid(unitID) or UnitInParty(unitID) then
        PollByUnitID(unitID, true)
    end
end

controller:SetScript("OnEvent", function(self, event, ...) self[event](self, ...) end)

-- api
controller.api = {}
ns.Inspect = controller.api

function controller.api.Enable()
    controller:RegisterEvent("INSPECT_READY")
    controller:RegisterEvent("GROUP_ROSTER_UPDATE")
    controller:RegisterEvent("PLAYER_SPECIALIZATION_CHANGED")
end

function controller.api.Disable()
    controller:UnregisterEvent("INSPECT_READY")
    controller:UnregisterEvent("GROUP_ROSTER_UPDATE")
    controller:UnregisterEvent("PLAYER_SPECIALIZATION_CHANGED")
end

function controller.api.filter(mode, ...)
    if mode == 'reset' then
        specFilter = setmetatable({}, { __index = Data.irrelevantSpec })
    end

    local numFlags = select('#', ...)
    for i=1, numFlags do
        local flag = select(i, ...)
        if flag then
            for k, v in pairs(Data.filterGroups[string.upper(flag)]) do
                specFilter[k] = mode and v or not v
            end
        end
    end
    shortcutFilter = {}
    for k, v in pairs(specFilter) do
        shortcutFilter[Data.specClass[k]] = shortcutFilter[Data.specClass[k]] or specFilter[k]
    end

    PollGroup(true)
end

dataobj.filter_func = controller.api.filter

function controller.api.TestPoll()
    Printer.debug(1, "Running test poll")
    PollByUnitID('player', true)
end

controller.api.state = setmetatable({ register = notifyRegister },
    { __index = function(t, v)
        if v == 'current' then
            return inspecting
        elseif v == 'script' then
            return controller:GetScript('OnUpdate') and 'set' or 'not set'
        end
    end })
